import inspect
from dataclasses import dataclass


@dataclass
class SourceLines:
    path: str
    start: int
    end: int

    def to_string(self) -> str:
        return f"{self.path}: {self.start}-{self.end}"


def get_source_info(cls: type) -> SourceLines:
    """Get the source file where the class is defined.

    Args:
        cls: The class to inspect.

    Returns:
        The source file path as a string.

    """
    try:
        source_path = inspect.getsourcefile(cls)
    except TypeError:
        return SourceLines("<built-in>", 0, 0)

    assert source_path is not None
    source_lines = inspect.getsourcelines(cls)
    start_line, end_line = source_lines[1], source_lines[1] + len(source_lines[0]) - 1
    return SourceLines(source_path, start_line, end_line)



from typing import Any

from .display import _display
from .merge import get_info


def defaults(
    *,
    include_dunders: bool | None = None,
    include_docs: bool | None = None,
    include_signatures: bool | None = None,
) -> None:
    ppdir.__kwdefaults__ = {
        "include_dunders": include_dunders,
        "include_docs": include_docs,
        "include_signatures": include_signatures,
    }


def ppdir(
    inp_cls: Any,
    *,
    include_dunders: bool = False,
    include_docs: bool = False,
    include_signatures: bool = False,
) -> None:
    if not isinstance(inp_cls, type):
        inp_cls = inp_cls.__class__
    class_summaries = get_info(inp_cls)
    try:
        _display(
            class_summaries,
            include_dunders=include_dunders,
            include_docs=include_docs,
            include_signatures=include_signatures,
        )
    except Exception as e:
        print(f"Error displaying info for {inp_cls}: {e}")
        print(dir(inp_cls))



import textwrap
from typing import TYPE_CHECKING

from colorama import Fore, Style, init

init()

if TYPE_CHECKING:
    from .merge import ClassSummary
INDENT = " " * 2
MAX_ALIGNING_LENGTH = 60


def is_dunder(name: str) -> bool:
    return name.startswith("__") and name.endswith("__")


def _display(
    data: list["ClassSummary"],
    *,
    include_dunders: bool = False,
    include_docs: bool = True,
    include_signatures: bool = False,
) -> None:
    # We reverse the class order so the most specific class is on the bottom since that's likely what people care about
    ret = ""
    for class_summary in data[::-1]:
        source_info = class_summary.source_info.to_string()
        ret += f"\n{Fore.BLUE}{class_summary.class_type.__name__} ({source_info}){Style.RESET_ALL}\n"

        if class_summary.class_type.__doc__ and include_docs:
            docstr = textwrap.indent(class_summary.class_type.__doc__.splitlines()[0], INDENT)
            ret += f"{Fore.GREEN}{docstr}{Style.RESET_ALL}\n\n"

        if class_summary.attr_info:
            ret += f"{INDENT}Attributes:\n"

            vals = sorted(class_summary.attr_info, key=lambda v: v.name)
            colon_position: int = max(val.colon_position() for val in vals)
            colon_position = colon_position if colon_position <= MAX_ALIGNING_LENGTH else 0

            for val in vals:
                if not include_dunders and is_dunder(val.name):
                    continue

                val_str = val.to_string(colon_position=colon_position, include_docs=include_docs)
                ret += f"{INDENT * 2}{Fore.GREEN}{val_str}{Style.RESET_ALL}\n"
        if class_summary.method_info:
            ret += f"{INDENT}Methods:\n"
            vals = sorted(class_summary.method_info, key=lambda v: v.name)
            colon_position: int = max(val.colon_position(include_signatures=include_signatures) for val in vals)
            colon_position = colon_position if colon_position <= MAX_ALIGNING_LENGTH else 0

            for val in vals:
                if not include_dunders and is_dunder(val.name):
                    continue

                val_str = val.to_string(
                    colon_position=colon_position,
                    include_docs=include_docs,
                    include_signatures=include_signatures,
                )
                ret += f"{INDENT * 2}{Fore.GREEN}{val_str}{Style.RESET_ALL}\n"

    print(ret)



def _expanded_dir(cls: type) -> list[str]:
    ret: list[str] = dir(cls)
    ret.extend(cls.__dict__.keys())
    if "__annotations__" in cls.__dict__:
        ret.extend(cls.__annotations__.keys())
    if "__pydantic_fields__" in cls.__dict__:
        # If a field exists, but isn't set by the instance, it doesn't appear in the dir() call
        ret.extend(cls.__pydantic_fields__.keys())
    return list(set(ret))


def get_class_defines(cls: type) -> dict[str, list[str]]:
    all_class_entities = _expanded_dir(cls)

    ret = {}
    for mro in cls.mro()[::-1]:
        ret[mro.__name__] = []
        for entity in _expanded_dir(mro):
            if entity in all_class_entities:
                ret[mro.__name__].append(entity)
                all_class_entities.remove(entity)
    return ret



import ast
import inspect
from dataclasses import dataclass
from typing import Literal


@dataclass
class AttrInfo:
    name: str
    type: str
    doc: str

    def _pre_colon_str(self) -> str:
        return f"{self.name} ({self.type})"

    def to_string(self, colon_position: int, *, include_docs: bool) -> str:
        pre_colon = f"{self._pre_colon_str()}:".ljust(colon_position + 1)

        if include_docs:
            doc_summary = self.doc.splitlines()[0] if self.doc else "--N/A--"
            return f"{pre_colon} {doc_summary}"

        return pre_colon

    def colon_position(self) -> int:
        return len(self._pre_colon_str())


@dataclass
class MethodInfo:
    name: str
    signature: str
    method_type: Literal["instance", "class", "static"]
    doc: str

    def _pre_colon_str(self, *, include_signatures: bool) -> str:
        pre_colon = self.name
        if include_signatures:
            pre_colon += f" ({self.signature})"
        return pre_colon

    def to_string(self, colon_position: int, *, include_docs: bool, include_signatures: bool) -> str:
        pre_colon = f"{self._pre_colon_str(include_signatures=include_signatures)}:".ljust(colon_position + 1)

        if self.method_type == "class":
            pre_colon = f"\b\bᶜ {pre_colon}"
        elif self.method_type == "static":
            pre_colon = f"\b\bˢ {pre_colon}"

        if include_docs:
            doc_summary = self.doc.splitlines()[0] if self.doc else "--N/A--"
            return f"{pre_colon} {doc_summary}"

        return pre_colon

    def colon_position(self, *, include_signatures: bool) -> int:
        return len(self._pre_colon_str(include_signatures=include_signatures))


def ast_find_classdef(tree: ast.Module) -> ast.ClassDef | None:
    for e in ast.walk(tree):
        if isinstance(e, ast.ClassDef):
            return e
    return None


def get_attr_docstrings(cls: type) -> list[AttrInfo]:
    try:
        src = inspect.getsource(cls)
    except TypeError:
        # This can occur when you try to get the source of a built-in, like dict
        return []

    tree = ast.parse(src)
    tree = ast_find_classdef(tree)
    assert tree is not None

    attribute_docs = []

    body = tree.body
    if not isinstance(body[0], ast.AnnAssign):
        body = body[1:]

    for expr in body:
        # When encouter an Expr, check if the expr a string
        if isinstance(expr, ast.Expr):
            if not attribute_docs:
                continue

            # The value is a ast.Value node
            # therefore another access to value is needed
            assert isinstance(expr.value, ast.Constant)
            doc_string = expr.value.value
            doc_string = doc_string.strip() if isinstance(doc_string, str) else None
            last_attr = attribute_docs[-1]
            if not last_attr.doc and doc_string is not None:
                last_attr.doc = doc_string

        # if the last known doc string is not none
        # and this next node is an annotation, that's a docstring
        if isinstance(expr, ast.AnnAssign):
            # expr.target is a ast.Name
            name = ast.unparse(expr.target)
            type_name = ast.unparse(expr.annotation)
            attribute_docs.append(
                AttrInfo(name=name, type=type_name, doc=""),
            )

    return attribute_docs


def get_method_docstrings(cls: type) -> list[MethodInfo]:
    methods = inspect.getmembers(cls, predicate=lambda x: inspect.ismethod(x) or inspect.isfunction(x))
    ret = []
    for method in methods:
        signature = inspect.signature(method[1])  # This is just to ensure it doesn't error

        method_with_possible_decorators = inspect.getattr_static(cls, method[0])
        method_type = "instance"
        if isinstance(method_with_possible_decorators, classmethod):
            method_type = "class"
        elif isinstance(method_with_possible_decorators, staticmethod):
            method_type = "static"

        ret.append(
            MethodInfo(
                name=method[0],
                signature=str(signature),
                method_type=method_type,
                doc=method[1].__doc__ or "",
            ),
        )
    return ret



from dataclasses import dataclass
from typing import Any

from ppdir import class_source_file

from .attr_doc_string import AttrInfo, MethodInfo, get_attr_docstrings, get_method_docstrings
from .get_class_defines import get_class_defines


@dataclass
class ClassSummary:
    class_type: type
    source_info: class_source_file.SourceLines
    attr_info: list[AttrInfo]
    method_info: list[MethodInfo]


def get_info(inp_cls: Any) -> list[ClassSummary]:
    class_defines = get_class_defines(inp_cls)
    ret = []
    for mro_cls in inp_cls.mro():
        source_info = class_source_file.get_source_info(mro_cls)

        defines = class_defines[mro_cls.__name__]
        attr_info = get_attr_docstrings(mro_cls)
        method_info = get_method_docstrings(mro_cls)
        filtered_attr_info = [v for v in attr_info if v.name in defines]
        filtered_method_info = [v for v in method_info if v.name in defines]
        ret.append(
            ClassSummary(
                class_type=mro_cls,
                source_info=source_info,
                attr_info=filtered_attr_info,
                method_info=filtered_method_info,
            ),
        )
    return ret
